package com.restlet.client.script.parser.impl;

import com.restlet.client.script.parser.ScriptParser;
import com.restlet.client.script.parser.token.ScriptToken;
import com.restlet.client.script.parser.token.ScriptTokenImpl;
import com.restlet.client.script.parser.token.ScriptTokenType;
import com.restlet.client.script.scanner.ScriptScanner;
import com.restlet.client.script.scanner.impl.ScriptScannerImpl;
import com.restlet.client.script.scanner.lex.ScriptLexType;
import com.restlet.client.script.scanner.lex.ScriptLexeme;
import com.restlet.client.utils.StringUtils;

import java.util.ArrayList;
import java.util.List;

public class ScriptParserImpl implements ScriptParser {

    private static final char EXPRESSION_START_CHARACTER = '{';
    private static final char EXPRESSION_END_CHARACTER = '}';

    public ScriptParserImpl() {}

    private static ScriptToken parseExpression(ScriptLexeme lex, final ScriptScanner scanner, final List<ScriptToken> tokens) {

        // expect '{'
        if (lex == null || !ScriptLexType.ExpressionBegin.equals(lex.getType())) {
            return null;
        }

        final ScriptTokenImpl token = new ScriptTokenImpl();
        token.setType(ScriptTokenType.Expression);
        token.setSourceIndex(lex.getIndex());

        final int scannerIndex = scanner.getIndex();

        lex = scanner.scan();

        // skip white spaces
        while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
            lex = scanner.scan();
        }

        if (lex == null) {
            scanner.setIndex(scannerIndex);
            return null;
        }

        final ScriptTokenImpl declToken = new ScriptTokenImpl();
        declToken.setType(ScriptTokenType.Statement);
        declToken.setSourceIndex(scannerIndex);

        final ScriptToken decl = parseStatement(declToken, lex, scanner);

        if (decl != null) {
            lex = scanner.scan();

            // skip white spaces
            while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
                lex = scanner.scan();
            }

            if (lex != null) {
                switch (lex.getType()) {
                    // expect '}'
                    case ExpressionEnd:
                        declToken.setSource(scanner.getSource(scannerIndex, lex.getIndex()));
                        token.append(decl);
                        token.setSource(scanner.getSource(token.getSourceIndex(), scanner.getIndex()));
                        return token;

                    default:
                        break;
                }
            }
        }

        // parse legacy context format
        scanner.setIndex(scannerIndex);

        final ScriptToken legacy = parseLegacyContext(scanner);

        if (legacy != null) {
            lex = scanner.scan();

            if (lex != null) {
                switch (lex.getType()) {
                    case ExpressionEnd:

                        token.append(legacy);
                        token.setSource(scanner.getSource(token.getSourceIndex(), scanner.getIndex()));
                        return token;

                    default:
                        break;
                }
            }
            scanner.setIndex(scannerIndex);
        }

        return null;
    }

    private static ScriptToken parseStatement(final ScriptTokenImpl declToken, ScriptLexeme lex, final ScriptScanner scanner) {

        if (lex == null) {
            return null;
        }

        // expect id or string
        final ScriptTokenImpl refToken = new ScriptTokenImpl();

        switch (lex.getType()) {
            case Id:
                refToken.setType(ScriptTokenType.Reference);
                break;

            case Number:
                refToken.setType(ScriptTokenType.Number);
                break;

            case QuotedString:
                refToken.setType(ScriptTokenType.String);
                break;

            default:
                return null;
        }

        final boolean id = ScriptLexType.Id.equals(lex.getType());
        refToken.setValue(lex.getValue());
        refToken.setSourceIndex(lex.getIndex());
        declToken.append(refToken);

        final int sourceIndex = scanner.getIndex();
        lex = scanner.scan();

        if (lex == null) {
            refToken.setSource(scanner.getSource(refToken.getSourceIndex(), scanner.getIndex()));
            scanner.setIndex(sourceIndex);
            return declToken;
        }

        switch (lex.getType()) {
            case BracketLeft:
                if (id) {
                    final ScriptTokenImpl argsToken = new ScriptTokenImpl();

                    do {

                        lex = scanner.scan();

                        // skip white spaces
                        while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
                            lex = scanner.scan();
                        }

                        // parse argument
                        if (lex != null && !ScriptLexType.BracketRight.equals(lex.getType())) {
                            final ScriptTokenImpl argToken = new ScriptTokenImpl();
                            argToken.setType(ScriptTokenType.Statement);
                            argToken.setSourceIndex(lex.getIndex());

                            final ScriptToken arg = parseStatement(argToken, lex, scanner);

                            if (arg != null) {
                                lex = scanner.scan();
                            }

                            argToken.setSource(scanner.getSource(argToken.getSourceIndex(), scanner.getIndex() - 1));
                            argsToken.append(argToken);

                            // skip white spaces
                            while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
                                lex = scanner.scan();
                            }
                        }

                        // expect ]
                        if (lex == null || !ScriptLexType.BracketRight.equals(lex.getType())) {
                            scanner.setIndex(sourceIndex);
                            return null;
                        }

                        lex = scanner.scan();

                        // skip white spaces
                        while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
                            lex = scanner.scan();
                        }

                    } while (lex != null && ScriptLexType.BracketLeft.equals(lex.getType()));

                    if (argsToken.getChildren() != null && !argsToken.getChildren().isEmpty()) {
                        for (final ScriptToken tk : argsToken.getChildren()) {
                            refToken.append(tk);
                        }
                        refToken.setType(ScriptTokenType.IndexReference);
                    } else {
                        refToken.setType(ScriptTokenType.Reference);
                    }

                }
                break;

            case ParenthesesLeft:
                if (id) {
                    lex = scanner.scan();

                    // skip white spaces
                    while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
                        lex = scanner.scan();
                    }

                    final ScriptTokenImpl argsToken = new ScriptTokenImpl();

                    // parse arguments
                    while (lex != null && !ScriptLexType.ParenthesesRight.equals(lex.getType())) {
                        final ScriptTokenImpl argToken = new ScriptTokenImpl();
                        argToken.setType(ScriptTokenType.Statement);
                        argToken.setSourceIndex(lex.getIndex());

                        final ScriptToken arg = parseStatement(argToken, lex, scanner);

                        if (arg != null) {
                            lex = scanner.scan();
                        }

                        argToken.setSource(scanner.getSource(argToken.getSourceIndex(), scanner.getIndex() - 1));
                        argsToken.append(argToken);

                        // skip white spaces
                        while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
                            lex = scanner.scan();
                        }

                        if (lex != null && ScriptLexType.Comma.equals(lex.getType())) {
                            lex = scanner.scan();
                            while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
                                lex = scanner.scan();
                            }

                        } else {
                            break;
                        }
                    }

                    // expect )
                    if (lex != null && ScriptLexType.ParenthesesRight.equals(lex.getType())) {
                        if (argsToken.getChildren() != null && !argsToken.getChildren().isEmpty()) {
                            for (final ScriptToken tk : argsToken.getChildren()) {
                                refToken.append(tk);
                            }
                            //argsToken.setSource(scanner.getSource(argsToken.getSourceIndex(), scanner.getIndex() - 1));
                        }
                        refToken.setType(ScriptTokenType.Method);
                        lex = scanner.scan();
                    } else {
                        scanner.setIndex(sourceIndex);
                        return null;
                    }
                }
                break;

            default:
                break;
        }

        if (lex == null) {
            scanner.setIndex(sourceIndex);
            return null;
        }

        final int endIndex = lex.getIndex();

        switch (lex.getType()) {
            // accept '.'
            case Dot:

                lex = scanner.scan();

                final ScriptToken nextToken = parseStatement(declToken, lex, scanner);
                if (nextToken != null) {
                    refToken.setSource(scanner.getSource(refToken.getSourceIndex(), endIndex));
                    return declToken;
                }
                break;

            default:
                refToken.setSource(scanner.getSource(refToken.getSourceIndex(), lex.getIndex()));
                scanner.setIndex(lex.getIndex());
                return declToken;
        }

        scanner.setIndex(sourceIndex);
        return null;
    }

    private static ScriptToken parseLegacyContext(final ScriptScanner scanner) {

        ScriptLexeme lex = scanner.scan();

        if (lex == null || ScriptLexType.ExpressionEnd.equals(lex.getType())) {
            return null;
        }

        final ScriptTokenImpl legacy = new ScriptTokenImpl();
        legacy.setType(ScriptTokenType.LegacyContext);
        legacy.setSourceIndex(lex.getIndex());

        scanner.setIndex(lex.getIndex());
        final ScriptLexeme ne = scanner.findExpressionBegin();

        if (ne != null) {
            scanner.setIndex(lex.getIndex());
            return null;
        }

        lex = scanner.scan();

        // skip white spaces
        while (lex != null && ScriptLexType.WhiteSpace.equals(lex.getType())) {
            lex = scanner.scan();
        }

        while (lex != null) {

            if (ScriptLexType.ExpressionEnd.equals(lex.getType())) {
                legacy.setSource(scanner.getSource(legacy.getSourceIndex(), lex.getIndex()));
                legacy.setValue(legacy.getSource());
                scanner.setIndex(lex.getIndex());
                return legacy;

            } else if (ScriptLexType.ExpressionBegin.equals(lex.getType())) {
                scanner.setIndex(lex.getIndex());
                return null;

            }
            lex = scanner.scan();
        }
        return null;
    }

    private static void addTextLex(final List<ScriptToken> tokens, final ScriptScanner scanner, final int start, final int end) {

        // join an existing text token
        if (!tokens.isEmpty() && ScriptTokenType.Text.equals(tokens.get(tokens.size() - 1).getType())

                ) {
            final ScriptTokenImpl token = (ScriptTokenImpl) tokens.get(tokens.size() - 1);

//			System.out.println("22: " + lex.getIndex() + ", " + scanner.getIndex());
            final String source = scanner.getSource(start, end);
//			System.out.println("23: " + source);

            if (token.getSource() != null) {
                token.setSource(token.getSource().concat(source));
            } else {
                token.setSource(source);
            }
            token.setSourceIndex(start);

            return;
        }

        // create a new text token
        final ScriptTokenImpl token = new ScriptTokenImpl();
        token.setType(ScriptTokenType.Text);
//		token.setValue(stringValue);
//		token.setSource(stringValue);
//		System.out.println("2: " + lex.getIndex() + ", " + scanner.getIndex());
        token.setSource(scanner.getSource(start, end));
//		System.out.println("3: " + token.getSource());
        token.setSourceIndex(start);
        tokens.add(token);
    }

    @Override
    public List<ScriptToken> parse(final String script) {

        if (StringUtils.isBlank(script)) {
            return null;
        }

        final ScriptScanner scanner = new ScriptScannerImpl(script);
        final List<ScriptToken> tokens = new ArrayList<>();

        // find an expression begin
        ScriptLexeme lex = scanner.findExpressionBegin();

        boolean foundExpression = false;

        while (lex != null) {
            switch (lex.getType()) {
                case ExpressionBegin:
                    final ScriptToken token = parseExpression(lex, scanner, tokens);
                    if (token != null) {
                        foundExpression = foundExpression || ScriptTokenType.Expression.equals(token.getType());
                        tokens.add(token);

                    } else {
                        addTextLex(tokens, scanner, lex.getIndex(), scanner.getIndex());
                    }
                    break;

                default:
                    addTextLex(tokens, scanner, lex.getIndex(), scanner.getIndex());
                    break;
            }
            lex = scanner.findExpressionBegin();

        }

        if (!foundExpression || tokens.isEmpty()) {
            return null;
        }

        // read remaining text
        if (!scanner.isEOF()) {
            addTextLex(tokens, scanner, scanner.getIndex(), scanner.getLength());
        }

//		printTokens(tokens, "T");

        return tokens.isEmpty() ? null : tokens;
    }

}
