package com.restlet.client.script.runtime;

import com.restlet.client.async.Promise;
import com.restlet.client.platform.CryptoApi;
import com.restlet.client.script.method.Base64Method;
import com.restlet.client.script.method.ConcatMethod;
import com.restlet.client.script.method.HMACMethod;
import com.restlet.client.script.method.HashMethod;
import com.restlet.client.script.method.JsonPathMethod;
import com.restlet.client.script.method.LengthMethod;
import com.restlet.client.script.method.LowerMethod;
import com.restlet.client.script.method.MD5Method;
import com.restlet.client.script.method.RandomNumberMethod;
import com.restlet.client.script.method.StringMethod;
import com.restlet.client.script.method.SubStringMethod;
import com.restlet.client.script.method.TimestampMethod;
import com.restlet.client.script.method.UUIDGeneratorMethod;
import com.restlet.client.script.method.UpperMethod;

import java.util.Collections;
import java.util.List;

import static com.restlet.client.utils.EnumUtils.getEnumNames;
import static java.util.Arrays.asList;
import static java.util.Collections.singletonList;

public interface ScriptMethod {

    String getName();

    Promise<ScriptValue> execute(ScriptValue value, List<ScriptValue> args);


    enum Type {
        // The first 4 are the default values
        base64(Base64Method.class, Collections.<ScriptMethodParameter>emptyList(), false),
        lower(LowerMethod.class, Collections.<ScriptMethodParameter>emptyList(), false),
        length(LengthMethod.class, Collections.<ScriptMethodParameter>emptyList(), false),
        substring(SubStringMethod.class, asList(
                new ScriptMethodParameter("Start", ScriptMethodParameter.Type.Integer, "0"),
                new ScriptMethodParameter("End", ScriptMethodParameter.Type.Integer)
        ), false),

        jsonPath(JsonPathMethod.class, singletonList(new ScriptMethodParameter("JSON path", ScriptMethodParameter.Type.String, "$")), false),

        // Crypto
        hmac(HMACMethod.class, asList(
                new ScriptMethodParameter("Cipher", getEnumNames(CryptoApi.HashType.values())),
                new ScriptMethodParameter("Secret", ScriptMethodParameter.Type.String),
                new ScriptMethodParameter("Output", getEnumNames(CryptoApi.Output.values()), CryptoApi.Output.Base64.name())
        ), false),
        sha(HashMethod.class, asList(
                new ScriptMethodParameter("Cipher", getEnumNames(CryptoApi.HashType.values())),
                new ScriptMethodParameter("Output", getEnumNames(CryptoApi.Output.values()), CryptoApi.Output.Base64.name())
        ), false),
        md5(MD5Method.class, Collections.<ScriptMethodParameter>emptyList(), false),

        // String
        upper(UpperMethod.class, Collections.<ScriptMethodParameter>emptyList(), false),
        string(StringMethod.class, singletonList(new ScriptMethodParameter("Quotes", asList("'", "\""), "\"")), false),
        concat(ConcatMethod.class, singletonList(new ScriptMethodParameter("Content", ScriptMethodParameter.Type.String)), false),

        // Utility
        timestamp(TimestampMethod.class, Collections.<ScriptMethodParameter>emptyList(), true),
        uuid(UUIDGeneratorMethod.class, Collections.<ScriptMethodParameter>emptyList(), true),
        random(RandomNumberMethod.class,
                singletonList(new ScriptMethodParameter("Maximum", ScriptMethodParameter.Type.Integer)),
                true);

        final boolean standalone;
        Class<? extends ScriptMethod> clazz;
        List<ScriptMethodParameter> parameters;

        Type(final Class<? extends ScriptMethod> clazz, final List<ScriptMethodParameter> parameters, final boolean standalone) {
            this.clazz = clazz;
            this.parameters = parameters;
            this.standalone = standalone;
        }

        public static Type getByName(final String name) {
            for (final Type type : values()) {
                if (type.name().equals(name)) {
                    return type;
                }
            }
            throw new RuntimeException("ScriptMethod type " + name + " does not exist.");
        }

        public List<ScriptMethodParameter> getParameters() {
            return parameters;
        }

        public boolean isStandalone() {
            return standalone;
        }

    }

}
