package com.restlet.client.script.scanner.impl;

import com.restlet.client.script.scanner.ScriptScanner;
import com.restlet.client.script.scanner.lex.ScriptLexType;
import com.restlet.client.script.scanner.lex.ScriptLexeme;
import com.restlet.client.script.scanner.lex.SimpleLexeme;
import com.restlet.client.script.scanner.lex.StringLexeme;
import com.restlet.client.utils.StringUtils;

public class ScriptScannerImpl implements ScriptScanner {

    final String script;
    final char[] chars;
    int index;

    public ScriptScannerImpl(final String script) {
        this.script = script;
        this.index = 0;
        this.chars = (script != null) ? script.toCharArray() : null;
    }

    static final boolean isAlpha(final char ch) {
        return
                ((ch >= 'a') && (ch <= 'z'))
                        || ((ch >= 'A') && (ch <= 'Z'))
                ;
    }

    static final boolean isNumber(final char ch) {
        return
                ((ch >= '0') && (ch <= '9'))
                ;
    }

    static final boolean isAlphaNum(final char ch) {
        return isAlpha(ch) || isNumber(ch);
    }

    static final boolean isWS(final char ch) {
        return ch == ' ' || ch == '\r' || ch == '\n' || ch == '\t';
    }

    @Override
    public ScriptLexeme scan() {
        if (isEOF()) {
            return null;
        }

        State state = State.Initial;

        final StringBuilder value = new StringBuilder();

        int i = index;

        while (i < chars.length) {

            final char ch = chars[i];

//			System.out.println("> " + ch + ", index: " + i + ", " + state + ", " + value);

            switch (state) {
                case Initial:
                    // id
                    if (isAlpha(ch) || ch == '_') {
                        if (value.length() > 0) {
                            value.append(ch);
                        } else {
                            value.append(ch);
                            state = State.Identifier;
                        }

                        // number
                    } else if (isNumber(ch)) {
                        if (value.length() > 0) {
                            value.append(ch);
                        } else {
                            value.append(ch);
                            state = State.Number;
                        }

                        // double quoted string
                    } else if (ch == '"') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        state = State.QuotedString;

                        // quoted string
                    } else if (ch == '\'') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        state = State.QuotedString;

                        // dot
                    } else if (ch == '.') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.Dot, value, index);

//				// dot
//				} else if (ch == '+') {
//					if (value.length() > 0) {
//						return createLexeme(ExpressionLexType.Text, value, index);
//					}
//					value.append(ch);
//					return createLexeme(ExpressionLexType.Plus, value, index);

                        // comma
                    } else if (ch == ',') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.Comma, value, index);

                        // double bracket open
                    } else if (ch == '{') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.ExpressionBegin, value, index);

                        // double bracket close
                    } else if (ch == '}') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.ExpressionEnd, value, index);

                        // parentheses left
                    } else if (ch == '(') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.ParenthesesLeft, value, index);

                        // parentheses right
                    } else if (ch == ')') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.ParenthesesRight, value, index);

                        // dollar sign
                    } else if (ch == '$') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        state = State.DollarExpression;

                    } else if (ch == '[') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.BracketLeft, value, index);

                    } else if (ch == ']') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.BracketRight, value, index);

                        // white space
                    } else if (isWS(ch)) {
                        if (!StringUtils.isBlank(value.toString())) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }

                        value.append(ch);

                        if ((i + 1) < chars.length && isWS(chars[i + 1])) {
                            // eat whitespaces

                        } else {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }

                        // text
                    } else {
                        value.append(ch);
                    }
                    break;

                case Identifier:
                    if (isAlphaNum(ch) || ch == '_') {
                        value.append(ch);

                    } else if (isWS(ch) || ch == '.' || ch == '}' || ch == '{' || ch == '(' || ch == ')' || ch == ']' || ch == '[' || ch == ',' /*|| ch == '+'*/) {
                        return createLexeme(ScriptLexType.Id, value, index);

                    } else {
                        value.append(ch);
                        state = State.Initial;
                    }
                    break;

                case Number:
                    if (isNumber(ch)) {
                        value.append(ch);

                    } else if (isWS(ch) || ch == '.' || ch == '}' || ch == '{' || ch == '(' || ch == ')' || ch == ']' || ch == '[' || ch == ',' /*|| ch == '+'*/) {
                        return createLexeme(ScriptLexType.Number, value, index);

                    } else {
                        value.append(ch);
                        state = State.Initial;
                    }
                    break;

                case QuotedString:
                    if (value.charAt(0) == ch) {
                        value.append(ch);
                        return createStringLexeme(value, index);

                    } else if (ch == '\\') {
                        state = State.EscapedQuote;

                    } else {
                        value.append(ch);
                    }
                    break;

                case DollarExpression:
                    if (ch == '{') {
                        value.append(ch);
                        return createLexeme(ScriptLexType.ExpressionBegin, value, index);
                    } else {
                        value.append(ch);
                        state = State.Initial;
                    }
                    break;

                case EscapedQuote:
                    if (ch == '\\') {
                        value.append('\\');
                        break;
                    }

                    if (value.charAt(0) != ch) {
                        value.append('\\');
                    }

                    value.append(ch);
                    state = State.QuotedString;
                    break;
            }
            i += 1;
        }

        switch (state) {
            case Identifier:
                return createLexeme(ScriptLexType.Id, value, index);
            case Number:
                return createLexeme(ScriptLexType.Number, value, index);
            default:
                return createLexeme(ScriptLexType.Text, value, index);
        }
    }

    @Override
    public ScriptLexeme findExpressionBegin() {
        if (isEOF()) {
            return null;
        }

        State state = State.Initial;

        final StringBuilder value = new StringBuilder();

        int i = index;

        while (i < chars.length) {

            final char ch = chars[i];

            switch (state) {
                case Initial:
                    // open double bracket
                    if (ch == '{') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append(ch);
                        return createLexeme(ScriptLexType.ExpressionBegin, value, index);

                        // dollar sign
                    } else if (ch == '$') {
                        state = State.DollarExpression;

                        // text
                    } else {
                        value.append(ch);
                    }
                    break;

                case DollarExpression:
                    if (ch == '{') {
                        if (value.length() > 0) {
                            return createLexeme(ScriptLexType.Text, value, index);
                        }
                        value.append('$');
                        value.append(ch);
                        return createLexeme(ScriptLexType.ExpressionBegin, value, index);

                    } else {
                        value.append(ch);
                        state = State.Initial;
                    }
                    break;

                default:
                    throw new IllegalStateException();
            }
            i += 1;
        }
        return null;
    }

    final ScriptLexeme createLexeme(final ScriptLexType type, final StringBuilder value, final int index) {

        final String stringValue = value.toString();

        ScriptLexeme lex = null;

        if (ScriptLexType.Text.equals(type) && StringUtils.isBlank(stringValue)) {
            lex = new SimpleLexeme(ScriptLexType.WhiteSpace, index, stringValue);

        } else {
            lex = new SimpleLexeme(type, index, stringValue);
        }

        this.index += stringValue.length();

        return lex;
    }

    final ScriptLexeme createStringLexeme(final StringBuilder value, final int index) {
        final String quotedString = value.toString();
        String stringValue = "";

        if (quotedString.length() > 2) {
            stringValue = quotedString.substring(1, quotedString.length() - 1);
        }

        final StringLexeme lex = new StringLexeme(stringValue, index, quotedString.charAt(0));

        int i = stringValue.indexOf(lex.getQuote());    // hack to find escaped quotes
        while (i != -1) {
            this.index += 1;
            i = stringValue.indexOf(lex.getQuote(), i + 1);
        }

        this.index += quotedString.length();

        return lex;
    }

    @Override
    public final boolean isEOF() {
        return (chars == null || (chars.length == 0) || (index >= chars.length));
    }

    @Override
    public ScriptLexeme close() {
        if (isEOF()) {
            return null;
        }

        final StringBuilder value = new StringBuilder();
        value.append(chars, index, chars.length - index);

        return createLexeme(ScriptLexType.Text, value, index);
    }

    @Override
    public String getSource(final int start, final int end) {
//		System.out.println(": " + start + "-" + end);
        if (start >= 0 && end > 0 && start < end && start < chars.length && end <= chars.length) {
            return String.copyValueOf(chars, start, end - start);
        }
        return null;
    }

    @Override
    public int getIndex() {
        return index;
    }

    @Override
    public void setIndex(final int index) {
        this.index = index;
    }

    @Override
    public int getLength() {
        return chars != null ? chars.length : 0;
    }

    enum State {
        DollarExpression,
        QuotedString,
        Identifier,
        EscapedQuote,
        Initial,
        Number,
    }
}
