package com.restlet.client.tests.asserts;

import com.restlet.client.function.Function;
import com.restlet.client.tests.SingleRequestTestResult;
import com.restlet.client.utils.Iterables;
import com.restlet.client.utils.Objects;
import com.restlet.client.utils.Sequence;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashSet;
import java.util.List;

import static com.restlet.client.tests.asserts.AssertionResultTo.AssertionState.None;
import static com.restlet.client.tests.asserts.AssertionResultTo.AssertionState.getStateComparator;
import static java.util.Arrays.asList;

public class AssertionResultTo implements AssertionTo {

    private AssertionState assertionState = None;
    private final AssertionTo assertion;
    private final Collection<AssertionResultMessage> messages = new LinkedHashSet<>();
    private final List<Throwable> errors = new ArrayList<>();
    private String evaluatedPath;
    private String evaluatedValue;
    private String actualValue;

    public AssertionResultTo(final AssertionTo assertion) {
        this.assertion = assertion;
    }

    @Override
    public Boolean isEnabled() {
        return assertion.isEnabled();
    }

    @Override
    public void setEnabled(final Boolean enabled) {
        assertion.setEnabled(enabled);
    }

    @Override
    public Subject getSubject() {
        return assertion.getSubject();
    }

    @Override
    public void setSubject(final Subject subject) {
        assertion.setSubject(subject);
    }

    @Override
    public String getPath() {
        return assertion.getPath();
    }

    @Override
    public void setPath(final String path) {
        assertion.setPath(path);
    }

    @Override
    public Comparison getComparison() {
        return assertion.getComparison();
    }

    @Override
    public void setComparison(final Comparison comparison) {
        assertion.setComparison(comparison);
    }

    @Override
    public String getValue() {
        return assertion.getValue();
    }

    @Override
    public void setValue(final String value) {
        assertion.setValue(value);
    }

    public final AssertionTo getAssertion() {
        return assertion;
    }

    public AssertionState getAssertionState() {
        return assertionState;
    }

    public void setAssertionState(final AssertionState assertionState) {
        this.assertionState = assertionState;
    }

    public Collection<AssertionResultMessage> getMessages() {
        return messages;
    }

    public Collection<AssertionResultMessage> getMessages(final AssertionState assertionState) {
        final Collection<AssertionResultMessage> messagesByState = new LinkedHashSet<>();
        for (final AssertionResultMessage message : messages) {
            if (message.assertionState == assertionState) {
                messagesByState.add(message);
            }
        }
        return messagesByState;
    }

    public void addFailureMessage(final String message) {
        messages.add(new AssertionResultMessage(AssertionState.Failure, message));
    }

    public void addErrorMessage(final String message) {
        messages.add(new AssertionResultMessage(AssertionState.Error, message));
    }

    public void addWarningMessage(final String message) {
        messages.add(new AssertionResultMessage(AssertionState.Warning, message));
    }

    public String getEvaluatedPath() {
        return evaluatedPath;
    }

    public void setEvaluatedPath(final String evaluatedPath) {
        this.evaluatedPath = evaluatedPath;
    }

    public String getEvaluatedValue() {
        return evaluatedValue;
    }

    public void setEvaluatedValue(final String evaluatedValue) {
        this.evaluatedValue = evaluatedValue;
    }

    public String getActualValue() {
        return actualValue;
    }

    public AssertionResultTo setActualValue(final String actualValue) {
        this.actualValue = actualValue;
        return this;
    }

    public List<Throwable> getErrors() {
        return errors;
    }

    public AssertionResultTo addError(final Throwable error) {
        this.errors.add(error);
        return this;
    }

    public void reset() {
        setActualValue(null);
        setAssertionState(None);
        setEvaluatedPath(null);
        setEvaluatedValue(null);
    }

    public enum AssertionState {
        Ok(SingleRequestTestResult.State.Ok),
        Failure(SingleRequestTestResult.State.Failure),
        Error(SingleRequestTestResult.State.Failure),
        Warning(SingleRequestTestResult.State.Ok),
        None(SingleRequestTestResult.State.None);

        private static final List<AssertionState> ORDERED_ASSERTION_STATES = asList(Error, Failure, Warning, Ok, None);

        public final SingleRequestTestResult.State testResultState;

        AssertionState(final SingleRequestTestResult.State testResultState) {
            this.testResultState = testResultState;
        }

        public static Comparator<AssertionState> getStateComparator() {
            return new Comparator<AssertionState>() {
                @Override
                public int compare(final AssertionState assertionState1, final AssertionState assertionState2) {
                    if (!ORDERED_ASSERTION_STATES.contains(assertionState1)) {
                        throw new RuntimeException("State " + assertionState1 + " is not in the ordered list of states.");
                    }
                    if (!ORDERED_ASSERTION_STATES.contains(assertionState2)) {
                        throw new RuntimeException("State " + assertionState2 + " is not in the ordered list of states.");
                    }
                    return Integer.compare(ORDERED_ASSERTION_STATES.indexOf(assertionState1), ORDERED_ASSERTION_STATES.indexOf(assertionState2));
                }
            };
        }
    }

    public static AssertionState getPrimaryState(final List<AssertionResultTo> assertionResults, final AssertionState defaultAssertionState) {
        final List<AssertionState> orderedAssertionStates = Sequence.of(assertionResults)
                .omitNulls()
                .map(new Function<AssertionResultTo, AssertionState>() {
                    @Override
                    public AssertionState apply(final AssertionResultTo assertionResult) {
                        return assertionResult.getAssertionState();
                    }
                })
                .omitNulls()
                .toList();
        Collections.sort(orderedAssertionStates, getStateComparator());

        return Iterables.getFirst(orderedAssertionStates)
                .orElse(defaultAssertionState);
    }

    public static class AssertionResultMessage {
        public AssertionState assertionState;
        public String message;

        AssertionResultMessage(final AssertionState assertionState, final String message) {
            this.assertionState = assertionState;
            this.message = message;
        }

        @Override
        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final AssertionResultMessage that = (AssertionResultMessage) o;

            return assertionState == that.assertionState
                    && Objects.equals(message, that.message);
        }

        @Override
        public int hashCode() {
            int result = assertionState != null ? assertionState.hashCode() : 0;
            result = 31 * result + (message != null ? message.hashCode() : 0);
            return result;
        }
    }
}
