package com.restlet.client.tests.asserts;

import com.restlet.client.function.Predicate;
import com.restlet.client.utils.Objects;
import com.restlet.client.utils.Sequence;

import java.util.List;

import static com.restlet.client.tests.asserts.AssertionTo.Comparison.Contains;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.DoesNotContain;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.DoesNotEqual;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.DoesNotExist;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.Equals;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.Exists;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.GreateOrEqual;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.Greater;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.Less;
import static com.restlet.client.tests.asserts.AssertionTo.Comparison.LessOrEqual;
import static com.restlet.client.tests.asserts.AssertionTo.Subject.Response;
import static com.restlet.client.tests.asserts.AssertionTo.Subject.ResponseBody;
import static com.restlet.client.tests.asserts.AssertionTo.Subject.ResponseStatus;

/**
 * An assertion is attached to a request.
 * But the __evaluation__ of the assertion is based on
 * the response (associated to the request).
 */
public interface AssertionTo {

    Boolean isEnabled();

    void setEnabled(Boolean enabled);

    Subject getSubject();

    void setSubject(Subject subject);

    String getPath();

    void setPath(String path);

    Comparison getComparison();

    void setComparison(Comparison comparison);

    String getValue();

    void setValue(String value);

    enum SubjectAndPath {
        ResponseStatusCode(ResponseStatus, "code", "Status code", Equals, DoesNotEqual, Less, LessOrEqual, Greater, GreateOrEqual),
        ResponseHeader(Subject.ResponseHeader, null, "Header", Comparison.values()),
        ResponseJsonBody(Subject.ResponseJsonBody, null, "JSON Body", Comparison.values()),
        ResponseXmlBody(Subject.ResponseXmlBody, null, "XML Body", Comparison.values()),
        ResponseBodyContent(ResponseBody, "content", "Body content", Equals, DoesNotEqual, Exists, DoesNotExist, Contains, DoesNotContain),
        ResponseBodyLength(ResponseBody, "length", "Body length", Equals, DoesNotEqual, Less, LessOrEqual, Greater, GreateOrEqual),
        RequestDurationInMs(Response, "latency", "Duration (ms)", Less, LessOrEqual, Greater, GreateOrEqual),
        ResponseStatusMessage(ResponseStatus, "message", "Status message", Equals, DoesNotEqual, Exists, DoesNotExist, Contains, DoesNotContain);

        public final Subject subject;
        public final String path;
        private final List<Comparison> comparisons;
        public final String label;

        SubjectAndPath(final Subject subject, final String path, final String label, final Comparison... comparisons) {
            this.subject = subject;
            this.path = path;
            this.label = label;
            this.comparisons = Sequence.of(comparisons).toList();
        }

        public List<Comparison> getComparisons() {
            return comparisons;
        }

        public static SubjectAndPath valueOf(final Subject subject, final String path) {
            return Sequence.of(SubjectAndPath.values())
                    .filter(new Predicate<SubjectAndPath>() {
                        @Override
                        public boolean test(final SubjectAndPath subjectAndPath) {
                            return subjectAndPath.subject == subject && (subjectAndPath.path == null || Objects.equals(subjectAndPath.path, path));
                        }
                    })
                    .first()
                    .get();
        }

    }

    enum Subject {
        ResponseStatus("Status", "the status", false),
        ResponseHeader("Header", "the header", true),
        ResponseJsonBody("JSON Body", "the JSON body identified by", false),
        ResponseXmlBody("XML Body", "the XML body identified by ", false),
        ResponseBody("Body", "the body", false),
        Response("Response", "the response", false);

        public final String dropdownText;
        public final String textualDescription;
        public final boolean supportExpressionBuilder;

        Subject(final String dropdownText, final String textualDescription, final boolean supportExpressionBuilder) {
            this.dropdownText = dropdownText;
            this.textualDescription = textualDescription;
            this.supportExpressionBuilder = supportExpressionBuilder;
        }

        public String asString() {
            return dropdownText;
        }

        public List<Comparison> getComparisons(final String path) {
            return SubjectAndPath.valueOf(this, path).getComparisons();
        }
    }

    enum Comparison {
        Equals(true, "Equals", "should be equal to"),
        DoesNotEqual(true, "Does not equal", "should not be equal to"),
        Exists(false, "Exists", "should exist"),
        DoesNotExist(false, "Does not exist", "should not exist"),
        Contains(true, "Contains", "should contain"),
        DoesNotContain(true, "Does not contain", "should not contain"),
        Less(true, "Less than", "should be less than"),
        LessOrEqual(true, "Less than or equal", "should be less than or equal to"),
        Greater(true, "Greater than", "should be greater than"),
        GreateOrEqual(true, "Greater than or equal", "should be greater than or equal to");

        private final boolean hasParameter;
        public final String dropdownName;
        public final String textualDescription;

        Comparison(final boolean hasParameter, final String dropdownName, final String textualDescription) {
            this.hasParameter = hasParameter;
            this.dropdownName = dropdownName;
            this.textualDescription = textualDescription;
        }

        public boolean hasParameter() {
            return hasParameter;
        }

        public String asString() {
            return dropdownName;
        }
    }

}
