package com.restlet.client.tests.impl;

import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;
import com.restlet.client.model.EntityTo;
import com.restlet.client.tests.ContainerTestResult;
import com.restlet.client.tests.SingleRequestTestResult;
import com.restlet.client.tests.TestResult;
import com.restlet.client.tests.impl.ExecutionStateStore.RunStatistics;
import com.restlet.client.tests.impl.ExecutionStateStore.RunType;
import com.restlet.client.utils.EntityUtils;
import com.restlet.client.utils.Objects;
import com.restlet.client.utils.Sequence;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import static com.restlet.client.tests.TestResult.ContainerStateComputationPolicy.STOP_ON_FAILURE;
import static com.restlet.client.tests.TestResult.State.InProgress;

public class ContainerTestResultImpl implements ContainerTestResult {

    private State state;

    private final List<TestResult> testResults = new ArrayList<>();

    private String name;

    private String fullName;

    private ContainerTestResult parent;

    private EntityTo entity;

    private long elapsedTime = 0L;

    private final ExecutionStateStore executionStateStore;

    private RunStatistics runStatistics;

    public ContainerTestResultImpl(final ExecutionStateStore executionStateStore) {
        this.executionStateStore = executionStateStore;
        this.state = InProgress;
    }

    @Override
    public State getState() {
        return state;
    }

    @Override
    public ContainerTestResultImpl setState(final State state) {
        this.state = state;
        return this;
    }

    @Override
    public List<TestResult> getTestResults() {
        return testResults;
    }

    @Override
    public <T extends TestResult> ContainerTestResultImpl setTestResults(final List<T> testResults) {
        this.testResults.clear();
        this.testResults.addAll(testResults);
        return this;
    }

    @Override
    public <T extends TestResult> ContainerTestResultImpl addTestResult(final T testResult) {
        this.testResults.add(testResult);
        return this;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public ContainerTestResultImpl setName(final String name) {
        this.name = name;
        return this;
    }

    @Override
    public String getFullName() {
        return fullName;
    }

    @Override
    public ContainerTestResultImpl setFullName(final String fullName) {
        this.fullName = fullName;
        return this;
    }

    @Override
    public ContainerTestResult getParent() {
        return parent;
    }

    @Override
    public ContainerTestResult setParent(final ContainerTestResult parent) {
        this.parent = parent;
        return this;
    }

    @Override
    public EntityTo getEntity() {
        return entity;
    }

    @Override
    public ContainerTestResult setEntity(final EntityTo entity) {
        this.entity = entity;
        return this;
    }

    @Override
    public long getElapsedTime() {
        return elapsedTime;
    }

    @Override
    public ContainerTestResult setElapsedTime(final long elapsedTime) {
        this.elapsedTime = elapsedTime;
        return this;
    }

    @Override
    public void refreshState(final RunType runType, final boolean fireEvent) {
        final State originalState = getState();
        final Set<State> states = Sequence.of(testResults)
                .map(new Function<TestResult, State>() {
                    @Override
                    public State apply(final TestResult testResult) {return testResult.getState();}
                })
                .toSet();

        final ContainerStateComputationPolicy containerStateComputationPolicy = getEntity().getType().containerStateComputationPolicy;
        final State newState = containerStateComputationPolicy.computeContainerState(states);
        setState(newState);

        final boolean allChildrenAchieved = Sequence.of(testResults)
                .all(new Predicate<TestResult>() {
                    @Override
                    public boolean test(final TestResult childTestResult) {
                        return childTestResult.getState().isTerminal;
                    }
                });
        // make sure that all children have been achieved, since this event handler is called every time the terminal status is updated

        final boolean isContainerTerminated = !originalState.isTerminal && newState.isTerminal &&
                (allChildrenAchieved || containerStateComputationPolicy == STOP_ON_FAILURE);
        if (isContainerTerminated && fireEvent) {
            runStatistics.onEnd(state);
            executionStateStore.onContainerEnded(this, runType);
        }

        if (parent != null) {
            parent.refreshState(runType, fireEvent);
        }
    }

    @Override
    public int computeNumberOfDynamicRequests() {
        return Sequence.of(testResults)
                .reduce(new BiFunction<Integer, TestResult, Integer>() {
                    @Override
                    public Integer apply(final Integer seed, final TestResult testResult) {
                        if (testResult instanceof ContainerTestResult) {
                            return seed + ((ContainerTestResult) testResult).computeNumberOfDynamicRequests();
                        }

                        return seed + (((SingleRequestTestResult) testResult).isDynamicRequest() ? 1 : 0);
                    }
                }, 0);
    }

    @Override
    public int computeNumberOfLocalRequests() {
        return Sequence.of(testResults)
                .reduce(new BiFunction<Integer, TestResult, Integer>() {
                    @Override
                    public Integer apply(final Integer seed, final TestResult testResult) {
                        if (testResult instanceof ContainerTestResult) {
                            return seed + ((ContainerTestResult) testResult).computeNumberOfLocalRequests();
                        }

                        return seed + (((SingleRequestTestResult) testResult).isLocalRequest() ? 1 : 0);
                    }
                }, 0);
    }

    @Override
    public void setRunStatistics(final RunStatistics runStatistics) {
        this.runStatistics = runStatistics;
    }

    @Override
    public RunStatistics getRunStatistics() {
        return this.runStatistics;
    }

    @Override
    public ContainerTestResult computeNameAndFullName() {
        this.name = entity.getName();
        this.fullName = EntityUtils.generateDottedName(entity);
        return this;
    }

    @Override
    public ContainerTestResult walk(final Consumer<TestResult> visitor) {
        visitor.consume(this);
        Sequence.of(testResults).each(new BiConsumer<TestResult, Integer>() {
            @Override
            public void consume(final TestResult child, final Integer index) {
                child.walk(visitor);
            }
        });
        return this;
    }

    @Override
    public boolean matches(final EntityTo entity) {
        return matches(entity.getId());
    }

    @Override
    public boolean matches(final String entityId) {
        return Objects.equals(entityId, getEntity().getId());
    }
}
