package com.restlet.client.tests.impl;

import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.model.EntityTo;
import com.restlet.client.model.EntityTreeNode;
import com.restlet.client.model.environment.EnvironmentTo;
import com.restlet.client.net.http.response.HttpResponseTo;
import com.restlet.client.platform.UUIDGenerator;
import com.restlet.client.tests.ContainerTestResult;
import com.restlet.client.tests.SingleRequestTestResult;
import com.restlet.client.tests.SingleRequestTryTestResult;
import com.restlet.client.tests.TestResult;
import com.restlet.client.tests.TestResult.State;
import com.restlet.client.utils.Objects;
import com.restlet.client.utils.Sequence;

import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import static com.restlet.client.model.EntityTo.Type.Request;
import static com.restlet.client.tests.TestResult.State.InProgress;
import static com.restlet.client.tests.TestResult.State.Invalidated;
import static com.restlet.client.tests.TestResult.State.Ok;
import static com.restlet.client.utils.FunctionalUtils.ifNotNull;
import static com.restlet.client.utils.Iterables.getLast;

public class ExecutionStateStore {

    public enum ResponsePreviewStatus {
        Visible, Hidden
    }

    public enum RunType {
        ALL(true),
        DRAFT(false),
        DB(true);

        public boolean updateStateInSpTree;

        RunType(final boolean updateStateInSpTree) {
            this.updateStateInSpTree = updateStateInSpTree;
        }
    }

    private final Map<String, SingleRequestTestResult> requestResultIndexedById = new HashMap<>();

    private final Map<String, ContainerTestResult> containerTestResultIndexedById = new HashMap<>();

    public final Map<String, ResponsePreviewStatus> responsePreviewStatusById = new HashMap<>();

    private final LinkedHashMap<String, BiConsumer<SingleRequestTestResult, RunType>> requestStartedHandlers = new LinkedHashMap<>();

    private final LinkedHashMap<String, BiConsumer<SingleRequestTryTestResult, RunType>> requestTryStartedHandlers = new LinkedHashMap<>();

    private final LinkedHashMap<String, BiConsumer<ContainerTestResult, RunType>> containerStartedHandlers = new LinkedHashMap<>();

    private final LinkedHashMap<String, BiConsumer<SingleRequestTestResult, RunType>> requestEndedHandlers = new LinkedHashMap<>();

    private final LinkedHashMap<String, BiFunction<SingleRequestTryTestResult, RunType, Promise<Void>>> requestTryEndedBlockingHandlers = new LinkedHashMap<>();

    private final LinkedHashMap<String, BiConsumer<ContainerTestResult, RunType>> containerEndedHandlers = new LinkedHashMap<>();

    private final LinkedHashMap<String, BiConsumer<EntityTo, RunType>> invalidatedHandlers = new LinkedHashMap<>();

    private final LinkedHashMap<String, BiConsumer<TestResult, RunType>> elapsedTimeUpdatedHandlers = new LinkedHashMap<>();

    private EntityTo entityRunning;

    public void index(final SingleRequestTestResult singleRequestTestResult) {
        Objects.requireNonNull(singleRequestTestResult);
        requestResultIndexedById.put(singleRequestTestResult.getEntity().getId(), singleRequestTestResult);
    }

    public void index(final ContainerTestResult containerTestResult, final EnvironmentTo environment) {
        Objects.requireNonNull(containerTestResult);
        final String entityId = containerTestResult.getEntity().getId();
        containerTestResult.setRunStatistics(new RunStatistics(environment));

        containerTestResultIndexedById.put(entityId, containerTestResult);
    }

    public HandlerRegistration addRequestStartedHandler(final BiConsumer<SingleRequestTestResult, RunType> singleRequestResultHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.requestStartedHandlers.put(handlerRegistrationId, singleRequestResultHandler);
        return createHandlerRegistration(requestStartedHandlers, handlerRegistrationId);
    }

    public void removeRequestStartedHandler(final String handlerRegistrationId) {
        this.requestStartedHandlers.remove(handlerRegistrationId);
    }

    public HandlerRegistration addRequestTryStartedHandler(final BiConsumer<SingleRequestTryTestResult, RunType> singleRequestTryResultHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.requestTryStartedHandlers.put(handlerRegistrationId, singleRequestTryResultHandler);
        return createHandlerRegistration(requestTryStartedHandlers, handlerRegistrationId);
    }

    public void removeRequestTryStartedHandler(final String handlerRegistrationId) {
        this.requestTryStartedHandlers.remove(handlerRegistrationId);
    }

    public HandlerRegistration addContainerStartedHandler(final BiConsumer<ContainerTestResult, RunType> containerResultHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.containerStartedHandlers.put(handlerRegistrationId, containerResultHandler);
        return createHandlerRegistration(containerStartedHandlers, handlerRegistrationId);
    }

    public void removeContainerStartedHandler(final String handlerRegistrationId) {
        this.containerStartedHandlers.remove(handlerRegistrationId);
    }

    public HandlerRegistration addRequestEndedHandler(final BiConsumer<SingleRequestTestResult, RunType> singleRequestResultHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.requestEndedHandlers.put(handlerRegistrationId, singleRequestResultHandler);
        return createHandlerRegistration(requestEndedHandlers, handlerRegistrationId);
    }

    public void removeRequestEndedHandler(final String handlerRegistrationId) {
        this.requestEndedHandlers.remove(handlerRegistrationId);
    }

    public HandlerRegistration addRequestTryEndedHandler(final BiConsumer<SingleRequestTryTestResult, RunType> singleRequestTryResultHandler) {
        return addRequestTryEndedHandler(new BiFunction<SingleRequestTryTestResult, RunType, Promise<Void>>() {
            @Override
            public Promise<Void> apply(final SingleRequestTryTestResult singleRequestTryTestResult, final RunType runType) {
                singleRequestTryResultHandler.consume(singleRequestTryTestResult, runType);
                return Promises.of();
            }
        });
    }

    public void removeRequestTryEndedHandler(final String handlerRegistrationId) {
        this.requestTryEndedBlockingHandlers.remove(handlerRegistrationId);
    }

    public HandlerRegistration addRequestTryEndedHandler(final BiFunction<SingleRequestTryTestResult, RunType, Promise<Void>> singleRequestTryResultHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.requestTryEndedBlockingHandlers.put(handlerRegistrationId, singleRequestTryResultHandler);
        return createHandlerRegistration(requestTryEndedBlockingHandlers, handlerRegistrationId);
    }

    public HandlerRegistration addContainerEndedHandler(final BiConsumer<ContainerTestResult, RunType> containerResultHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.containerEndedHandlers.put(handlerRegistrationId, containerResultHandler);
        return createHandlerRegistration(containerEndedHandlers, handlerRegistrationId);
    }

    public HandlerRegistration addInvalidatedHandler(final BiConsumer<EntityTo, RunType> invalidatedHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.invalidatedHandlers.put(handlerRegistrationId, invalidatedHandler);
        return createHandlerRegistration(invalidatedHandlers, handlerRegistrationId);
    }

    public HandlerRegistration addElapsedTimeUpdatedHandler(final BiConsumer<TestResult, RunType> elapsedTimeHandler) {
        final String handlerRegistrationId = UUIDGenerator.uuid();
        this.elapsedTimeUpdatedHandlers.put(handlerRegistrationId, elapsedTimeHandler);
        return createHandlerRegistration(elapsedTimeUpdatedHandlers, handlerRegistrationId);
    }

    void onRequestStarted(final SingleRequestTestResult singleRequestTestResult, final RunType runType) {
        Sequence.of(requestStartedHandlers.values())
                .each(new BiConsumer<BiConsumer<SingleRequestTestResult, RunType>, Integer>() {
                    @Override
                    public void consume(final BiConsumer<SingleRequestTestResult, RunType> handler, final Integer index) {
                        handler.consume(singleRequestTestResult, runType);
                    }
                });
    }

    void onRequestTryStarted(final SingleRequestTryTestResult singleRequestTryTestResult, final RunType runType) {
        Sequence.of(requestTryStartedHandlers.values())
                .each(new BiConsumer<BiConsumer<SingleRequestTryTestResult, RunType>, Integer>() {
                    @Override
                    public void consume(final BiConsumer<SingleRequestTryTestResult, RunType> handler, final Integer index) {
                        handler.consume(singleRequestTryTestResult, runType);
                    }
                });
    }

    void onContainerStarted(final ContainerTestResult containerTestResult, final RunType runType) {
        Sequence.of(containerStartedHandlers.values())
                .each(new BiConsumer<BiConsumer<ContainerTestResult, RunType>, Integer>() {
                    @Override
                    public void consume(final BiConsumer<ContainerTestResult, RunType> handler, final Integer index) {
                        handler.consume(containerTestResult, runType);
                    }
                });
    }

    void onRequestEnded(final SingleRequestTestResult singleRequestTestResult, final RunType runType) {
        Sequence.of(requestEndedHandlers.values())
                .each(new BiConsumer<BiConsumer<SingleRequestTestResult, RunType>, Integer>() {
                    @Override
                    public void consume(final BiConsumer<SingleRequestTestResult, RunType> handler, final Integer index) {
                        handler.consume(singleRequestTestResult, runType);
                    }
                });
    }

    Promise<Void> onRequestTryEnded(final SingleRequestTryTestResult singleRequestTryTestResult, final RunType runType) {
        return Promises.all(Sequence.of(requestTryEndedBlockingHandlers.values())
                .map(new Function<BiFunction<SingleRequestTryTestResult, RunType, Promise<Void>>, Promise<Void>>() {
                    @Override
                    public Promise<Void> apply(final BiFunction<SingleRequestTryTestResult, RunType, Promise<Void>> blockingHandler) {
                        return blockingHandler.apply(singleRequestTryTestResult, runType);
                    }
                })
                .toList())
                .castToVoidPromise();
    }

    void onContainerEnded(final ContainerTestResult containerTestResult, final RunType runType) {
        Sequence.of(containerEndedHandlers.values())
                .each(new BiConsumer<BiConsumer<ContainerTestResult, RunType>, Integer>() {
                    @Override
                    public void consume(final BiConsumer<ContainerTestResult, RunType> handler, final Integer index) {
                        handler.consume(containerTestResult, runType);
                    }
                });
    }

    private void onInvalidated(final EntityTo entity, final RunType runType, final boolean dropLastResponse) {
        if (containerTestResultIndexedById.containsKey(entity.getId())) {
            containerTestResultIndexedById.get(entity.getId()).refreshState(RunType.ALL, false);
        }

        if (requestResultIndexedById.containsKey(entity.getId())) {
            if (dropLastResponse) {
                requestResultIndexedById.remove(entity.getId());
            } else {
                requestResultIndexedById.get(entity.getId()).setState(Invalidated);
            }
        }

        Sequence.of(invalidatedHandlers.values())
                .each(new BiConsumer<BiConsumer<EntityTo, RunType>, Integer>() {
                    @Override
                    public void consume(final BiConsumer<EntityTo, RunType> handler, final Integer secondValue) {
                        handler.consume(entity, runType);
                    }
                });
    }

    void onElapsedTimeUpdated(final TestResult testResult, final RunType runType) {
        Sequence.of(elapsedTimeUpdatedHandlers.values())
                .each(new BiConsumer<BiConsumer<TestResult, RunType>, Integer>() {
                    @Override
                    public void consume(final BiConsumer<TestResult, RunType> handler, final Integer index) {
                        handler.consume(testResult, runType);
                    }
                });
    }

    public EntityTo getEntityRunning() {
        return entityRunning;
    }

    ExecutionStateStore setEntityRunning(final EntityTo entityRunning) {
        this.entityRunning = entityRunning;
        return this;
    }

    public SingleRequestTestResult getLastRequestResult(final String requestId) {
        return requestResultIndexedById.get(requestId);
    }

    public HttpResponseTo getResponse(final String requestId) {
        return ifNotNull(requestResultIndexedById.get(requestId), new Function<SingleRequestTestResult, HttpResponseTo>() {
            @Override
            public HttpResponseTo apply(final SingleRequestTestResult result) {
                final List<SingleRequestTryTestResult> triesResults = result.getTriesResults();
                for (final SingleRequestTryTestResult tryResult : triesResults) {
                    if (tryResult.getState() == Ok) {
                        return tryResult.getHttpResult().getResponse();
                    }
                }
                return getLast(triesResults).getHttpResult().getResponse();
            }
        });
    }

    public ContainerTestResult getLastContainerResult(final String containerId) {
        return containerTestResultIndexedById.get(containerId);
    }

    private void recursiveClean(final EntityTreeNode treeNode, final RepositoryDao repositoryDao, final boolean dropLastResponse) {
        // invalidate node and its children
        treeNode.walk(new Consumer<EntityTreeNode>() {
            @Override
            public void consume(final EntityTreeNode node) {
                ExecutionStateStore.this.onInvalidated(node.entity, RunType.ALL, dropLastResponse);
            }
        });

        // invalidate ancestors of the node
        EntityTo entity = treeNode.entity;

        while (entity != null && !entity.getType().isDrive) {
            final EntityTo persistedEntity = repositoryDao.readSync(entity.getId());
            final EntityTo parent = persistedEntity.getParent();

            if (parent == null) {
                break;
            }

            ExecutionStateStore.this.onInvalidated(parent, RunType.ALL, dropLastResponse);

            entity = parent;
        }
    }

    public void recursiveCleanAndDropLastResponse(final EntityTreeNode treeNode, final RepositoryDao repositoryDao) {
        recursiveClean(treeNode, repositoryDao, true);
    }

    public void recursiveCleanAndKeepLastResponse(final EntityTreeNode treeNode, final RepositoryDao repositoryDao) {
        recursiveClean(treeNode, repositoryDao, false);
    }

    private boolean isContainerRunning(final String containerId) {
        return containerTestResultIndexedById.containsKey(containerId)
                && containerTestResultIndexedById.get(containerId).getState() == InProgress;
    }

    public boolean isContainerRunning(final EntityTo entity) {
        return entity != null && isContainerRunning(entity.getId());
    }

    private boolean isRequestRunning(final String requestId) {
        return requestResultIndexedById.containsKey(requestId)
                && requestResultIndexedById.get(requestId).getState() == InProgress;
    }

    public boolean isRequestRunning(final EntityTo entity) {
        return entity != null && entity.getType() == Request && isRequestRunning(entity.getId());
    }

    public static class RunStatistics {
        private final Date startDate = new Date();
        public final String environmentName;

        private long durationInMS = 0;
        public State result = State.InProgress;

        private RunStatistics(final EnvironmentTo environment) {
            this.environmentName = environment == null ? "None" : environment.getName();
        }

        void onEnd(final State finalState) {
            result = finalState;
            durationInMS = new Date().getTime() - startDate.getTime();
        }

        public String getDuration() {
            if (durationInMS == 0) {
                return "-";
            }
            return ((durationInMS / 10) / 100.0) + "s";
        }

        public Date getStartDate() {
            return startDate;
        }
    }

    private static HandlerRegistration createHandlerRegistration(final LinkedHashMap<String, ?> cache, final String id) {
        return new HandlerRegistration() {
            @Override
            public void removeHandler() {
                cache.remove(id);
            }
        };
    }

    public interface HandlerRegistration {
        void removeHandler();
    }
}
