package com.restlet.client.utils;

import com.restlet.client.function.Predicate;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import static com.restlet.client.utils.Objects.firstNonNull;

public class CollectionUtils {

    private static <T> EqualityFunction<T> objectEquals() {
        return new EqualityFunction<T>() {
            @Override
            public boolean equals(final T a, final T b) {
                return Objects.equals(a, b);
            }
        };
    }

    public static <T> CollectionDiff<T> diff(final Collection<T> left, final Collection<T> right) {
        return new CollectionDiff<>(left, right, CollectionUtils.<T>objectEquals());
    }

    public static <T> CollectionDiff<T> diff(final List<T> left, final List<T> right) {
        return new CollectionDiff<>(left, right, CollectionUtils.<T>objectEquals());
    }

    public static <T> CollectionDiff<T> diff(final Collection<T> left, final Collection<T> right,
                                             final EqualityFunction<T> equalityFunction) {
        return new CollectionDiff<>(left, right, equalityFunction);
    }

    public static <T> CollectionDiff<T> diff(final List<T> left, final List<T> right,
                                             final EqualityFunction<T> equalityFunction) {
        return new CollectionDiff<>(left, right, equalityFunction);
    }

    public interface EqualityFunction<T> extends Serializable {
        boolean equals(T left, T right);
    }

    public static class CollectionPatchConfiguration<T> {

        public final SpecialEqualsList<T> toBeAdded;
        public final SpecialEqualsList<T> toBeRemoved;
        public final SpecialEqualsList<T> toBeUpdated;

        public CollectionPatchConfiguration() {
            this(CollectionUtils.<T>objectEquals());
        }

        public CollectionPatchConfiguration(final EqualityFunction<T> equalityFunction) {
            toBeAdded = new SpecialEqualsList<>(equalityFunction);
            toBeRemoved = new SpecialEqualsList<>(equalityFunction);
            toBeUpdated = new SpecialEqualsList<>(equalityFunction);
        }

        public boolean addInToBeAdded(final T newElement) {
            return addIfNotPresent(toBeAdded, newElement);
        }

        public boolean addInToBeRemoved(final T newElement) {
            return addIfNotPresent(toBeRemoved, newElement);
        }

        public boolean addInToBeUpdated(final T newElement) {
            return addIfNotPresent(toBeUpdated, newElement);
        }

        private boolean addIfNotPresent(final SpecialEqualsList<T> list, final T newElement) {
            return !list.contains(newElement) && list.add(newElement);
        }
    }

    public static class CollectionDiff<T> {
        public final Set<T> onlyInLeft = new LinkedHashSet<>();
        public final Set<T> onlyInRight = new LinkedHashSet<>();
        public final Set<T> inBoth = new LinkedHashSet<>();

        private CollectionDiff(final Collection<T> left, final Collection<T> right, final EqualityFunction<T> equalityFunction) {
            final SpecialEqualsList<T> failSafeLeft = firstNonNull(new SpecialEqualsList<T>(left, equalityFunction), new SpecialEqualsList<T>(equalityFunction));
            final SpecialEqualsList<T> failSafeRight = firstNonNull(new SpecialEqualsList<T>(right, equalityFunction), new SpecialEqualsList<T>(equalityFunction));

            for (final T element : failSafeLeft) {
                if (failSafeRight.contains(element)) {
                    inBoth.add(element);
                } else {
                    onlyInLeft.add(element);
                }
            }

            onlyInRight.addAll(Sequence.of(failSafeRight)
                    .filter(new Predicate<T>() {
                        @Override
                        public boolean test(final T rightElement) {
                            return !failSafeLeft.contains(rightElement);
                        }
                    })
                    .toList());
        }

        public boolean hasNoDiff() {
            return onlyInLeft.isEmpty() && onlyInRight.isEmpty();
        }
    }

    private static class SpecialEqualsList<T> extends ArrayList<T> {

        EqualityFunction<T> equalityFunction;

        SpecialEqualsList(final Collection<? extends T> c, final EqualityFunction<T> equalityFunction) {
            super(c);
            this.equalityFunction = equalityFunction;
        }

        SpecialEqualsList(final EqualityFunction<T> equalityFunction) {
            super();
            this.equalityFunction = equalityFunction;
        }

        private SpecialEqualsList() {
        }

        @Override
        public int indexOf(final Object o) {
            if (o == null) {
                for (int i = 0; i < size(); i++) {
                    if (get(i) == null) {
                        return i;
                    }
                }
            } else {
                for (int i = 0; i < size(); i++) {
                    final T left;
                    try {
                        left = (T) o;
                    } catch (final Exception e) {
                        return -1;
                    }

                    if (equalityFunction.equals(left, get(i))) {
                        return i;
                    }
                }
            }
            return -1;
        }
    }
}
