package com.restlet.client.utils;

import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Executor;
import com.restlet.client.function.Function;
import com.restlet.client.function.Supplier;

import java.util.Collection;
import java.util.List;

public class FunctionalUtils {

    public static <R, T> T ifNotNull(final R object, final Function<R, T> getter) {
        return ifNotNull(object, getter, null);
    }

    public static <R, T> T ifNotNull(final R object, final Function<R, T> getter, final T fallback) {
        return object == null ? fallback : getter.apply(object);
    }

    public static <T> Function<T, T> identity() {
        return new Function<T, T>() {
            @Override
            public T apply(final T input) {
                return input;
            }
        };
    }

    public static <T> void doIfNotNull(final T object, final Consumer<T> successHandler) {
        doIfNotNull(object, successHandler, noopExecutor());
    }

    public static <T> void doIfNotNull(final T object, final Consumer<T> successHandler, final Executor failHandler) {
        doIf(object != null,
                new Executor() {
                    @Override
                    public void execute() {
                        successHandler.consume(object);
                    }
                }, failHandler);
    }

    public static void doIf(final boolean condition, final Executor successExecutor) {
        doIf(condition, successExecutor, noopExecutor());
    }

    public static void doIf(final boolean condition, final Executor successExecutor, final Executor failureExecutor) {
        if (condition) {
            successExecutor.execute();
        } else {
            failureExecutor.execute();
        }
    }

    public static <T> T doIfOrThrow(final boolean condition, final Supplier<T> successExecutor, final RuntimeException fallbackException) {
        if (condition) {
            return successExecutor.get();
        } else {
            throw fallbackException;
        }
    }

    public static <R, T> BiConsumer<R, T> noopBiConsumer() {
        return new BiConsumer<R, T>() {
            @Override
            public void consume(final R firstValue, final T secondValue) {}
        };
    }

    public static <T> Consumer<T> noopConsumer() {
        return new Consumer<T>() {
            @Override
            public void consume(final T value) {}
        };
    }

    public static Executor noopExecutor() {
        return new Executor() {
            @Override
            public void execute() {}
        };
    }

    public static <T> T returnIf(final boolean condition, final Supplier<T> successSupplier, final Supplier<T> failureSupplier) {
        if (condition) {
            return successSupplier.get();
        } else {
            return failureSupplier.get();
        }
    }

    public static <T> Pair<Maybe<T>, List<T>> headAndTail(final Collection<T> inputCollection) {
        final Maybe<T> head = Iterables.getFirst(inputCollection);
        final List<T> tail = Sequence.of(inputCollection)
                .dropLeft(1)
                .toList();

        return new Pair<>(head, tail);
    }

    public static <T> T firstNotNull(final T first, final T... others) {
        if (first != null) {
            return first;
        }

        for (final T element : others) {
            if (element != null) {
                return element;
            }
        }
        throw new IllegalArgumentException("One input at least should be non-null");
    }

    public static int size(final List<?> items) {
        if (items == null) {
            return 0;
        }
        return items.size();
    }

}
